//========================================================== IBJgrd.c
//
// Grid Functions
// ==============
//
// Copyright (C) Dr. Lutz Janicke, Dunum, Germany, 1999
// email: lj@janicke.de
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// history:
//
// last change: 98-07-31  lj
//
//==================================================================

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include "IBJmsg.h"
#include "IBJtxt.h"
#include "IBJmth.h"
#include "IBJary.h"
#include "IBJtbl.h"
#include "IBJgrd.h"

#define G         (*pGrd)
#define BP(i,j)   (*(float*)AryPtrX((pGrd) ? &(pGrd->bpdsc) : NULL,(i),(j)))
#define BM(i,j)   (*(float*)AryPtrX((pGrd) ? &(pGrd->bmdsc) : NULL,(i),(j)))
#define SET(n)    if (n<0||n>1||!Grids[n].flag) {vMsg("Grid %d undefined!\n",(n));\
                    exit(1);} pGrd=Grids+(n)

static char *eMODn = "IBJgrd";

/*-----------------------------------------------------------------------------*/

GRDDEFREC *pGrd;

static GRDDEFREC Grids[2];

/*========================================================================= GrdInit
*/
int GrdInit( void )
  {
  pGrd = Grids;
  return 0;
  }

static void ClcH(float hv[4], float wv[4], float *ph, float *pw )
  {
  float wl, wr, hm, hl, hr;
  if (wv[1]==0 || wv[2]==0) {
    *pw = 0;
    *ph = 0;
    return;
    }
  hm = 0.5*(hv[1] + hv[2]);
  if (wv[0] > 0) {
    hl = hv[1] + 0.5*(hv[1]-hv[0]);
    wl = 1;
    }
  else { hl = 0;  wl = 0; }
  if (wv[3] > 0) {
    hr = hv[2] + 0.5*(hv[2]-hv[3]);
    wr = 1;
    }
  else { hr = 0;  wr = 0; }
  *ph = 0.5*((3*hm+hl)/(3+wl) + (3*hm+hr)/(3+wr));
  *pw = 1;
  return;
  }

static float ClcHeight( float zz[4][4], float ww[4][4] )
  {
  dP(ClcHeight);
  int i, j;
  float h, hxv[4], hyv[4], wxv[4], wyv[4], hv[4], wv[4];
  float hx, wx, hy, wy;
  for (i=0; i<4; i++) {
    for (j=0; j<4; j++) {
      hv[j] = zz[i][j];
      wv[j] = ww[i][j];
      }
    ClcH(hv, wv, hxv+i, wxv+i);
    }
  ClcH(hxv, wxv, &hx, &wx);
  for (j=0; j<4; j++) {
    for (i=0; i<4; i++) {
      hv[i] = zz[i][j];
      wv[i] = ww[i][j];
      }
    ClcH(hv, wv, hyv+j, wyv+j);
    }
  ClcH(hyv, wyv, &hy, &wy);
  if (wx==0 && wy==0)               eX(1);
  h = (wx*hx + wy*hy)/(wx + wy);
  return h;
eX_1:
  vMsg("undefined height! (%s)\n", ePGMs);
  return 0; 
  }

static int MakeBm( void )
  {
  dP(MakeBm);
  int sz, i, j, is, js, ii, jj;
  float zz[4][4], ww[4][4], z;
  sz = G.pbp->elmsz;
  AryCreate(G.pbm, sz, 2, 1, G.nx, 1, G.ny);    eG(1);
  for (i=1; i<=G.nx; i++)
    for (j=1; j<=G.ny; j++) {
      for (is=0; is<4; is++) 
        for (js=0; js<4; js++) {
          ii = i + is - 2;
          jj = j + js - 2;
          if (ii>=0 && ii<=G.nx && jj>=0 && jj<=G.ny) {
            zz[is][js] = BP(ii,jj);
            ww[is][js] = 1;
            }
          else {
            zz[is][js] = 0;
            ww[is][js] = 0;
            }
          }
      z = ClcHeight(zz, ww);                    eG(2);
      if (G.zh>0 && z>=G.zh)                    eX(3);
      BM(i,j) = z;
      }
  return 0;
eX_1:
  eMSG("can't create surface on interleaved grid! (%s)\n", ePGMs);
eX_2:
  eMSG("can't calculate height(%d,%d) on interleaved grid! (%s)\n", i, j, ePGMs);
eX_3:
  eMSG("surface height bm(%d,%d)=%e exceeds Zh=%e! (%s)\n", i, j, z, G.zh, ePGMs);
  }

/*==================================================================== GrdDefK
*/
int GrdDefK( int grdnum, float x0, float y0, float x1, float y1,
  float dd, int nz, float *ss, float sh, float zh )
  {
  dP(GrdDefK);
  int i, j, k;
  ARYDSC *pbp, *pbm;
  if (grdnum<0 || grdnum>1)             eX(1);
  if (dd <= 0)                          eX(2);
  if (nz < 1)                           eX(3);
  if (!ss)                              eX(4);
  if (ss[0] != 0)                       eX(5);
  for (k=1; k<=nz; k++)
    if (ss[k] <= ss[k-1])               eX(6);
  pGrd = Grids + grdnum;
  G.x0 = x0;
  G.y0 = y0;
  G.s0 = 0;
  G.dd = dd;
  G.nz = nz;
  G.ss = ss;
  G.s1 = ss[nz];
  if (zh > 0) { 
    G.sh = (sh > 0) ? sh : zh; 
    G.zh = zh; 
    }
  else { G.sh = 0;   G.zh = 0; }
  pbp = &G.bpdsc;
  pbm = &G.bmdsc;
  if (pbp) {
    if (!pbp->start)                                    eX(7);
    G.nx = pbp->bound[0].hgh;
    G.ny = pbp->bound[1].hgh;
    AryAssert(pbp, pbp->elmsz, 2, 0, G.nx, 0, G.ny);    eG(8);
    G.pbp = pbp;
    if (G.zh > 0) 
      for (i=0; i<=G.nx; i++)
        for (j=0; j<=G.ny; j++) 
          if (BP(i,j) >= G.zh)                          eX(20);
    if (x1 <= x0)  x1 = x0 + G.nx*dd;
    if (y1 <= y0)  y1 = y0 + G.ny*dd;
    if (!pbm)  pbm = &G.bmdsc;
    G.pbm = pbm;
    if (pbm->start) {
      AryAssert(pbm, pbp->elmsz, 2, 1, G.nx, 1, G.ny);  eG(11);
      }
    else {
      MakeBm();                                         eG(12);
      }
    }
  if (x1 <= x0)                         eX(13);
  if (y1 <= y0)                         eX(14);
  G.x1 = x1;
  G.y1 = y1;
  G.flag = 1;
  return 0;
eX_1: eX_2: eX_3: eX_4: eX_5: eX_6: eX_7: eX_8:
eX_11: eX_12: eX_13: eX_14:
  eMSG("invalid grid definition! (%s)\n", ePGMs);
eX_20:
  eMSG("surface height at (%d,%d) exceeds Zh=%e! (%s)\n", i, j, G.zh, ePGMs);
  }
  
/*==================================================================== GrdSet
*/
int GrdSet( int gn )
  {
  SET(gn);
  return 0;
  }
  
/*====================================================================== GrdI
*/
int GrdI( float x )
  {
  int i;
  if (x<G.x0 || x>G.x1)  return 0;
  i = 1 + (x-G.x0)/G.dd;
  if (i > G.nx)  i = G.nx;
  return i;
  }
  
/*====================================================================== GrdJ
*/
int GrdJ( float y )
  {
  int j;
  if (y<G.y0 || y>G.y1)  return 0;
  j = 1 + (y-G.y0)/G.dd;
  if (j > G.ny)  j = G.ny;
  return j;
  }
  
/*====================================================================== GrdK
*/
int GrdK( float s )
  {
  int k;
  if (s<G.s0 || s>G.s1)  return 0;
  for (k=1; k<=G.nz; k++)
    if (G.ss[k] > s)  break;
  if (k > G.nz)  k = G.nz;
  return k;
  }
  
/*==================================================================== GrdLoc2
*/
int GrdLoc2( float x, float y, 
int *pi, int *pj, float *pxi, float *peta, int *pl )
  {
  int i, j, flag;
  float xa, ya, xi, eta;
  
  flag = 0;
  if      (x < G.x0)  flag |= GRD_XLOW;
  else if (x > G.x1)  flag |= GRD_XHGH;
  if      (y < G.y0)  flag |= GRD_YLOW;
  else if (y > G.y1)  flag |= GRD_YHGH;
  if (flag)  return flag;
  
  i = 1 + (x-G.x0)/G.dd;
  if (i > G.nx)  i = G.nx;
  j = 1 + (y-G.y0)/G.dd;
  if (j > G.ny)  j = G.ny;
  if (pi)  *pi = i;
  if (pj)  *pj = j;
  if (!pxi && !peta && !pl)  return 0;
  xa = G.x0 + (i-1)*G.dd;
  ya = G.y0 + (j-1)*G.dd;
  xi  = (x - xa)/G.dd;
  eta = (y - ya)/G.dd;
  if (pxi)  *pxi = xi;
  if (peta)  *peta = eta;
  if (!pl)  return 0;
  if (xi >= eta)
    if (xi < 1-eta)  *pl = 1;
    else             *pl = 2;
  else
    if (xi < 1-eta)  *pl = 4;
    else             *pl = 3;
  return 0;
  }

/*==================================================================== GrdBij
*/
float GrdBij( int i, int j, float xi, float eta, int l )
  {
  dP(GrdBij);
  float b00, b01, b10, b11, bm, b;
  bm = BM(i,j);
  switch (l) {
    case 1: b00 = BP(i-1,j-1);
            b10 = BP(i,j-1);
            b = (1-xi)*b00 + xi*b10 + eta*(2*bm-b00-b10);
            break;
    case 2: b10 = BP(i,j-1);
            b11 = BP(i,j);
            b = (1-eta)*b10 + eta*b11 + (1-xi)*(2*bm-b10-b11);
            break;
    case 3: b01 = BP(i-1,j);
            b11 = BP(i,j);
            b = (1-xi)*b01 + xi*b11 + (1-eta)*(2*bm-b01-b11);
            break;
    case 4: b00 = BP(i-1,j-1);
            b01 = BP(i-1,j);
            b = (1-eta)*b00 + eta*b01 + xi*(2*bm-b00-b01);
            break;
    default:eX(1);
    }
  return b;
eX_1:
  vMsg("invalid subgrid number (%d)! (%s)\n", l, ePGMs);
  return 0;
  } 

/*==================================================================== GrdBxy
*/
float GrdBxy( float x, float y )
  {
  dP(GrdBxy);
  int i, j, l;
  float xi, eta, b;
  if (GrdLoc2(x, y, &i, &j, &xi, &eta, &l))     eX(1);
  b = GrdBij(i, j, xi, eta, l);                 eG(2);
  return b;
eX_1:
  vMsg("point (%e,%e) out of grid! (%s)\n", x, y, ePGMs);
  return 0;
eX_2:
  vMsg("can't locate (%e,%e)! (%s)\n", x, y, ePGMs);
  return 0;
  }
  
/*====================================================================== GrdS
*/
float GrdS( float x, float y, float z )
  {
  dP(GrdS);
  float b, s;
  b = GrdBxy(x, y);             eG(1);
  if (z < b)                    eX(2);
  if (G.sh > 0)  s = G.sh*(z-b)/(G.zh-b);
  else           s = z - b;
  return s;
eX_1:
  vMsg("can't calculate surface height! (%s)\n", ePGMs);
  return 0;
eX_2:  
  vMsg("z=%e below surface b=%e! (%s)\n", z, b, ePGMs);
  return 0;
  }
  
/*==================================================================== GrdClcS
*/
int GrdClcS( float x, float y, float *pz, float *ps )
  {
  dP(GrdClcS);
  float b, s, z;
  int rc;
  rc = 0;
  if (!pz)                                  eX(0);    
  z = *pz;
  b = GrdBxy(x, y);                         eG(1);
  if (z < b) {
    z = 2*b - z;  if (G.sh>0 && z>G.zh)     eX(2);
    *pz = z;
    rc = GRD_SLOW;
    }
  else if (G.sh>0 && z>G.zh) {
    z = 2*G.zh - z;  if (z < b)             eX(3);
    *pz = z;
    rc = GRD_SHGH;
    }
  if (G.sh > 0)  s = G.sh*(z-b)/(G.zh-b);
  else           s = z - b;
  if (ps)  *ps = s;
  return rc;
eX_0:
  eMSG("undefined parameter! (%s)\n", ePGMs);
eX_1:
  eMSG("can't calculate surface height! (%s)\n", ePGMs);
eX_2: eX_3: 
  eMSG("can't map z=%e into [%e,%e]! (%s)\n", z, b, G.zh, ePGMs);
  }
  
/*==================================================================== GrdOut
*/
int GrdOut( float x, float y, float s )
  {
  int flag;
  flag = 0;
  if      (x < G.x0)  flag |= GRD_XLOW;
  else if (x > G.x1)  flag |= GRD_XHGH;
  if      (y < G.y0)  flag |= GRD_YLOW;
  else if (y > G.y1)  flag |= GRD_YHGH;
  if      (s < G.s0)  flag |= GRD_SLOW;
  else if (s > G.s1)  flag |= GRD_SHGH;
  return flag;
  }

/*====================================================================== GrdZ
*/
int GrdZ( float x, float y, float s, float *pz )
  {
  dP(GrdZ);
  float b, z;
  int flag;
  flag = GrdOut(x, y, s);  if (flag)  return flag;
  b = GrdBxy(x, y);             eG(1);
  if (s < 0)                    eX(2);
  if (G.sh > 0)  z = b + s*(G.zh-b)/G.sh;
  else           z = b + s;
  if (pz)  *pz = z;
  return 0;
eX_1:
  eMSG("can't calculate surface height! (%s)\n", ePGMs);
eX_2:  
  eMSG("s=%e below surface! (%s)\n", s, ePGMs);
  }

/*==================================================================== GrdLoc3
*/
int GrdLoc3( float x, float y, float z, GRDLOCREC *ploc, int *pl )
  {
  dP(GrdLoc3);
  int i, j, k, l;
  float xi, eta, zeta, b, s;
  if (!ploc)  return 0;
  ploc->flag = GrdLoc2(x, y, &i, &j, &xi, &eta, &l);
  if (ploc->flag)  return ploc->flag;
  b = GrdBij(i, j, xi, eta, l);                 eG(2);
  if (G.sh > 0)  s = G.sh*(z-b)/(G.zh-b);
  else           s = z - b;
  if      (s < G.s0)  ploc->flag = GRD_SLOW;
  else if (s > G.s1)  ploc->flag = GRD_SHGH;
  if (ploc->flag)  return ploc->flag;
  for (k=1; k<=G.nz; k++)
    if (G.ss[k] > s)  break;
  if (k > G.nz)  k = G.nz;
  zeta = (s - G.ss[k-1])/(G.ss[k] - G.ss[k-1]);
  ploc->x = x;
  ploc->y = y;
  ploc->z = z;
  ploc->s = s;
  ploc->i = i;
  ploc->j = j;
  ploc->k = k;
  ploc->a = xi;
  ploc->b = eta;
  ploc->c = zeta;
  if (pl)  *pl = l;
  return 0;
eX_2:
  eMSG("can't locate (%e,%e,%e)! (%s)\n", x, y, z, ePGMs);
  }

/*===================================================================== GrdZp
*/
float GrdZp( int i, int j, int k )
  {
  dP(GrdZp);
  float b, z, s;
  if (i<0 || i>G.nx || j<0 || j>G.ny || k<0 || k>G.nz)  eX(1);
  b = BP(i,j);
  s = G.ss[k];
  if (G.sh > 0)  z = b + s*(G.zh-b)/G.sh;
  else           z = b + s;
  return z;
eX_1:
  vMsg("invalid index (%d,%d,%d)! (%s)\n", i, j, k, ePGMs);
  return 0;
  }
  
/*===================================================================== GrdZm
*/
float GrdZm( int i, int j, int k )
  {
  dP(GrdZm);
  float b, z, s;
  if (i<1 || i>G.nx || j<1 || j>G.ny || k<0 || k>G.nz)  eX(1);
  b = BM(i,j);
  s = G.ss[k];
  if (G.sh > 0)  z = b + s*(G.zh-b)/G.sh;
  else           z = b + s;
  return z;
eX_1:
  vMsg("invalid index (%d,%d,%d)! (%s)\n", i, j, k, ePGMs);
  return 0;
  }

/*===================================================================== GrdVp
*/
VECTOR GrdVp( int i, int j, int k )
  {
  dP(GrdVp);
  float x, y, z;
  VECTOR r;
  if (i<0 || i>G.nx || j<0 || j>G.ny || k<0 || k>G.nz)  eX(1);
  x = G.x0 + i*G.dd;
  y = G.y0 + j*G.dd;
  z = GrdZp(i,j,k);
  r = DefV(x, y, z);
  return r;
eX_1:
  vMsg("invalid index (%d,%d,%d)! (%s)\n", i, j, k, ePGMs);
  return DefV(0,0,0);
  }

/*===================================================================== GrdVm
*/
VECTOR GrdVm( int i, int j, int k )
  {
  dP(GrdVm);
  float x, y, z;
  VECTOR r;
  if (i<1 || i>G.nx || j<1 || j>G.ny || k<0 || k>G.nz)  eX(1);
  x = G.x0 + (i-0.5)*G.dd;
  y = G.y0 + (j-0.5)*G.dd;
  z = GrdZm(i,j,k);
  r = DefV(x, y, z);
  return r;
eX_1:
  vMsg("invalid index (%d,%d,%d)! (%s)\n", i, j, k, ePGMs);
  return DefV(0,0,0);
  }

/*================================================================= GrdS2Zij
*/
float GrdS2Zij( int i, int j )
  {
  dP(GrdS2Zij);
  float f;
  if (i<0 || i>G.nx || j<0 || j>G.ny )          eX(1);
  if (G.sh > 0)  f = (G.zh - BP(i,j))/G.sh;
  else           f = 1;
  return f;
eX_1:
  vMsg("invalid index (%d,%d)! (%s)\n", i, j, ePGMs);
  return 0;
  }
  
/*================================================================== GrdS2Z
*/
float GrdS2Z( float x, float y )
  {
  dP(GrdS2Z);
  float f;
  if (G.sh > 0) { f = (G.zh - GrdBxy(x,y))/G.sh;  eG(1); }
  else  f = 1;
  return f;
eX_1:
  vMsg("invalid position (%e,%e)! (%s)\n", x, y, ePGMs);
  return 0;
  }
    
/*===================================================================== GrdDp
*/
float GrdDp( int i, int j, int k )
  {
  dP(GrdDp);
  float d;
  if (i<0 || i>G.nx || j<0 || j>G.ny || k<1 || k>G.nz)  eX(1);
  d = G.ss[k] - G.ss[k-1];
  if (G.sh > 0)  d *= (G.zh - BP(i,j))/G.sh;
  return d;
eX_1:
  vMsg("invalid index (%d,%d,%d)! (%s)\n", i, j, k, ePGMs);
  return 0;
  }
  
/*===================================================================== GrdDm
*/
float GrdDm( int i, int j, int k )
  {
  dP(GrdDm);
  float d;
  if (i<1 || i>G.nx || j<1 || j>G.ny || k<1 || k>G.nz)      eX(1);
  d = G.ss[k] - G.ss[k-1];
  if (G.sh > 0)  d *= (G.zh - BM(i,j))/G.sh;
  return d;
eX_1:
  vMsg("invalid index (%d,%d,%d)! (%s)\n", i, j, k, ePGMs);
  return 0;
  }

/*==================================================================== GrdVol
*/
float GrdVol( int i, int j, int k )
  {
  dP(GrdVol);
  float d00, d01, d10, d11, dm, vol;
  if (i<1 || i>G.nx || j<1 || j>G.ny || k<1 || k>G.nz)      eX(1);
  d00 = GrdDp(i-1, j-1, k);                                 eG(2);
  d01 = GrdDp(i-1, j,   k);                                 eG(3);
  d10 = GrdDp(i  , j-1, k);                                 eG(4);
  d11 = GrdDp(i  , j,   k);                                 eG(5);
  dm  = GrdDm(i  , j,   k);                                 eG(6);
  vol = G.dd*G.dd*(d00 + d01 + d10 + d11 + 2*dm)/6;
  return vol;
eX_1: eX_2: eX_3: eX_4: eX_5: eX_6:
  vMsg("invalid index (%d,%d,%d)! (%s)\n", i, j, k, ePGMs);
  return 0;
  }

/*==================================================================== GrdArX
*/
float GrdArX( int i, int j, int k )
  {
  dP(GrdArX);
  float d10, d11, arx;
  if (i<0 || i>G.nx || j<1 || j>G.ny || k<1 || k>G.nz)      eX(1);
  d10 = GrdDp(i  , j-1, k);                                 eG(4);
  d11 = GrdDp(i  , j,   k);                                 eG(5);
  arx = 0.5*G.dd*(d10 + d11);
  return arx;
eX_1: eX_4: eX_5:
  vMsg("invalid index (%d,%d,%d)! (%s)\n", i, j, k, ePGMs);
  return 0;
  }

/*==================================================================== GrdArY
*/
float GrdArY( int i, int j, int k )
  {
  dP(GrdArY);
  float d01, d11, ary;
  if (i<1 || i>G.nx || j<0 || j>G.ny || k<1 || k>G.nz)      eX(1);
  d01 = GrdDp(i-1, j,   k);                                 eG(3);
  d11 = GrdDp(i  , j,   k);                                 eG(5);
  ary = 0.5*G.dd*(d01 + d11);
  return ary;
eX_1: eX_3: eX_5:
  vMsg("invalid index (%d,%d,%d)! (%s)\n", i, j, k, ePGMs);
  return 0;
  }

/*=================================================================== GrdNrmL
*/
VECTOR GrdNrmL( int i, int j, int k, int l )
  {
  dP(GrdNrmL);
  VECTOR a0, a1, a2, a3, a4, c, r00, r01, r10, r11, rm;
  rm = GrdVm(i, j, k);                              eG(1);
  switch (l) {
    case 1: r00 = GrdVp(i-1, j-1, k);               eG(2);
            r10 = GrdVp(i  , j-1, k);               eG(3);
            a0 = SubVV(r00, rm);
            a1 = SubVV(r10, rm);
            c = CroVV(a0, a1);
            break;
    case 2: r10 = GrdVp(i  , j-1, k);               eG(4);
            r11 = GrdVp(i  , j  , k);               eG(5);
            a1 = SubVV(r10, rm);
            a2 = SubVV(r11, rm);
            c = CroVV(a1, a2);
            break;
    case 3: r11 = GrdVp(i  , j  , k);               eG(6);
            r01 = GrdVp(i-1, j  , k);               eG(7);
            a2 = SubVV(r11, rm);
            a3 = SubVV(r01, rm);
            c = CroVV(a3, a2);
            break;
    case 4: r01 = GrdVp(i-1, j  , k);               eG(8);
            r00 = GrdVp(i-1, j-1, k);               eG(9);
            a3 = SubVV(r01, rm);
            a4 = SubVV(r00, rm);
            c = CroVV(a3, a4);
            break;
    default:eX(10);
    }
  c = MulSV(0.5, c);
  return c;
eX_1: eX_2: eX_3: eX_4: eX_5: eX_6: eX_7: eX_8: eX_9: eX_10:
  vMsg("invalid index (%d,%d,%d,%d)! (%s)\n", i, j, k, l, ePGMs);
  return DefV(0,0,0);
  }
  
/*==================================================================== GrdNrm
*/
int GrdNrm( int i, int j, int k, 
VECTOR *p1, VECTOR *p2, VECTOR *p3, VECTOR *p4 )
  {
  dP(GrdNrm);
  VECTOR a0, a1, a2, a3, a4, r00, r01, r10, r11, rm;
  rm  = GrdVm(i, j, k);                             eG(1);
  r00 = GrdVp(i-1, j-1, k);                         eG(2);
  r10 = GrdVp(i  , j-1, k);                         eG(3);
  r11 = GrdVp(i  , j  , k);                         eG(4);
  r01 = GrdVp(i-1, j  , k);                         eG(5);
  a0 = SubVV(r00, rm);
  a1 = SubVV(r10, rm);
  a2 = SubVV(r11, rm);
  a3 = SubVV(r01, rm);
  a4 = a0;
  if (p1)  *p1 = MulSV(0.5, CroVV(a0, a1));
  if (p2)  *p2 = MulSV(0.5, CroVV(a1, a2));
  if (p3)  *p3 = MulSV(0.5, CroVV(a2, a3));
  if (p4)  *p4 = MulSV(0.5, CroVV(a3, a4));
  return 0;
eX_1: eX_2: eX_3: eX_4: eX_5: 
  eMSG("invalid index (%d,%d,%d)! (%s)\n", i, j, k, ePGMs);
  }

/*===================================================================== GrdI3
*/
int GrdI3( GRDLOCREC loc, ARYDSC *pa, float *p )
  {
  dP(GrdI3); 
  int i, j, k, n, nn;
  float xi, eta, zeta, xis, etas, zetas;
  float *p000, *p001, *p010, *p011, *p100, *p101, *p110, *p111;
  if (!pa || !p)                                        eX(1);
  if (!pa->start)                                       eX(2);
  if (pa->numdm != 3)                                   eX(3);
  if (loc.flag)                                         eX(4);
  i = loc.i;
  j = loc.j;
  k = loc.k;
  xi   = loc.a;
  eta  = loc.b;
  zeta = loc.c;
  xis   = 1 - xi;
  etas  = 1 - eta;
  zetas = 1 - zeta;
  p000 = AryPtr(pa, i-1, j-1, k-1);  if (!p000)     eX(10);
  p001 = AryPtr(pa, i-1, j-1, k  );  if (!p001)     eX(11);
  p010 = AryPtr(pa, i-1, j  , k-1);  if (!p010)     eX(12);
  p011 = AryPtr(pa, i-1, j  , k  );  if (!p011)     eX(13);
  p100 = AryPtr(pa, i  , j-1, k-1);  if (!p100)     eX(14);
  p101 = AryPtr(pa, i  , j-1, k  );  if (!p101)     eX(15);
  p110 = AryPtr(pa, i  , j  , k-1);  if (!p110)     eX(16);
  p111 = AryPtr(pa, i  , j  , k  );  if (!p111)     eX(17);
  nn = pa->elmsz/sizeof(float);
  for (n=0; n<nn; n++) {
    p[n] = xis*etas*zetas*p000[n] + xi*etas*zetas*p100[n] + xis*eta*zetas*p010[n]
         + xis*etas*zeta*p001[n] + xi*eta*zetas*p110[n] + xi*etas*zeta*p101[n]
         + xis*eta*zeta*p011[n] + xi*eta*zeta*p111[n];
    }
  return 0;
eX_1: eX_2: eX_3: eX_4:
  eMSG("invalid parameter! (%s)\n", ePGMs);
eX_10: eX_11: eX_12: eX_13: eX_14: eX_15: eX_16: eX_17:
  eMSG("data not available at (%d,%d,%d)! (%s)\n", i, j, k, ePGMs);
  }
  
/*===================================================================== GrdI1X
*/
int GrdI1X( GRDLOCREC loc, ARYDSC *pa, float *p )
  {
  dP(GrdI1X); 
  int i, j, k;
  float xi, xis;
  float *p0, *p1;
  if (!pa || !p)                                        eX(1);
  if (!pa->start)                                       eX(2);
  if (pa->numdm != 3)                                   eX(3);
  if (loc.flag)                                         eX(4);
  i = loc.i;
  j = loc.j;
  k = loc.k;
  xi  = loc.a;
  xis = 1 - xi;
  p0 = AryPtr(pa, i-1, j, k);  if (!p0)     eX(10);
  p1 = AryPtr(pa, i  , j, k);  if (!p1)     eX(11);
  *p = xis*p0[0] + xi*p1[0];
  return 0;
eX_1: eX_2: eX_3: eX_4:
  eMSG("invalid parameter! (%s)\n", ePGMs);
eX_10: eX_11: 
  eMSG("data not available at (%d,%d,%d)! (%s)\n", i, j, k, ePGMs);
  }
  
/*===================================================================== GrdI1Y
*/
int GrdI1Y( GRDLOCREC loc, ARYDSC *pa, float *p )
  {
  dP(GrdI1Y); 
  int i, j, k;
  float eta, etas;
  float *p0, *p1;
  if (!pa || !p)                                        eX(1);
  if (!pa->start)                                       eX(2);
  if (pa->numdm != 3)                                   eX(3);
  if (loc.flag)                                         eX(4);
  i = loc.i;
  j = loc.j;
  k = loc.k;
  eta  = loc.b;
  etas = 1 - eta;
  p0 = AryPtr(pa, i, j-1, k);  if (!p0)     eX(10);
  p1 = AryPtr(pa, i, j  , k);  if (!p1)     eX(11);
  *p = etas*p0[0] + eta*p1[0];
  return 0;
eX_1: eX_2: eX_3: eX_4:
  eMSG("invalid parameter! (%s)\n", ePGMs);
eX_10: eX_11: 
  eMSG("data not available at (%d,%d,%d)! (%s)\n", i, j, k, ePGMs);
  }
  
/*===================================================================== GrdI1Z
*/
int GrdI1Z( GRDLOCREC loc, ARYDSC *pa, float *p )
  {
  dP(GrdI1Z); 
  int i, j, k;
  float zeta, zetas;
  float *p0, *p1;
  if (!pa || !p)                                        eX(1);
  if (!pa->start)                                       eX(2);
  if (pa->numdm != 3)                                   eX(3);
  if (loc.flag)                                         eX(4);
  i = loc.i;
  j = loc.j;
  k = loc.k;
  zeta  = loc.c;
  zetas = 1 - zeta;
  p0 = AryPtr(pa, i, j, k-1);  if (!p0)     eX(10);
  p1 = AryPtr(pa, i, j, k  );  if (!p1)     eX(11);
  *p = zetas*p0[0] + zeta*p1[0];
  return 0;
eX_1: eX_2: eX_3: eX_4:
  eMSG("invalid parameter! (%s)\n", ePGMs);
eX_10: eX_11: 
  eMSG("data not available at (%d,%d,%d)! (%s)\n", i, j, k, ePGMs);
  }
  
/*====================================================================== GrdMx
*/
int GrdMx( int i, int j, int k, ARYDSC *pa, float *p )
  {
  dP(GrdMx); 
  int n, nn;
  float *p00, *p01, *p10, *p11;
  if (!pa || !p)                                eX(1);
  if (!pa->start)                               eX(2);
  if (pa->numdm != 3)                           eX(3);
  nn = pa->elmsz/sizeof(float);
  p00 = AryPtr(pa, i, j-1, k-1);  if (!p00)     eX(10);
  p01 = AryPtr(pa, i, j-1, k  );  if (!p01)     eX(11);
  p10 = AryPtr(pa, i, j  , k-1);  if (!p10)     eX(12);
  p11 = AryPtr(pa, i, j  , k  );  if (!p11)     eX(13);
  for (n=0; n<nn; n++) 
    p[n] = 0.25*(p00[n] + p01[n] + p10[n] + p11[n]);
  return 0;
eX_1: eX_2: eX_3:
  eMSG("invalid parameter! (%s)\n", ePGMs);
eX_10: eX_11: eX_12: eX_13:
  eMSG("data not available at (%d,%d,%d)! (%s)\n", i, j, k, ePGMs);
  }

/*====================================================================== GrdMy
*/
int GrdMy( int i, int j, int k, ARYDSC *pa, float *p )
  {
  dP(GrdMy); 
  int n, nn;
  float *p00, *p01, *p10, *p11;
  if (!pa || !p)                                eX(1);
  if (!pa->start)                               eX(2);
  if (pa->numdm != 3)                           eX(3);
  nn = pa->elmsz/sizeof(float);
  p00 = AryPtr(pa, i-1, j, k-1);  if (!p00)     eX(10);
  p01 = AryPtr(pa, i-1, j, k  );  if (!p01)     eX(11);
  p10 = AryPtr(pa, i  , j, k-1);  if (!p10)     eX(12);
  p11 = AryPtr(pa, i  , j, k  );  if (!p11)     eX(13);
  for (n=0; n<nn; n++) 
    p[n] = 0.25*(p00[n] + p01[n] + p10[n] + p11[n]);
  return 0;
eX_1: eX_2: eX_3:
  eMSG("invalid parameter! (%s)\n", ePGMs);
eX_10: eX_11: eX_12: eX_13:
  eMSG("data not available at (%d,%d,%d)! (%s)\n", i, j, k, ePGMs);
  }

/*===================================================================== GrdMz4
*/
int GrdMz4( int i, int j, int k, ARYDSC *pap, ARYDSC *pam,
float *p1, float *p2, float *p3, float *p4 )
  {
  dP(GrdMz4); 
  int n, nn;
  float *p00, *p01, *p10, *p11, *pm;
  float w9[9], *wa;
  if (!pap)                                      eX(1);
  if (!pap->start)                               eX(2);
  if (pap->numdm != 3)                           eX(3);
  nn = pap->elmsz/sizeof(float);
  p00 = AryPtr(pap, i-1, j-1, k);  if (!p00)     eX(10);
  p01 = AryPtr(pap, i-1, j  , k);  if (!p01)     eX(11);
  p10 = AryPtr(pap, i  , j-1, k);  if (!p10)     eX(12);
  p11 = AryPtr(pap, i  , j  , k);  if (!p11)     eX(13);
  wa = NULL;
  if (pam) {
    pm = AryPtr(pam, i, j, k);  if (!pm)         eX(14);
    }
  else
    if (nn > 9) {
      wa = ALLOC(nn*sizeof(float));  if (!wa)    eX(15);
      pm = wa;
      }
    else {
      pm = w9;
      }
  for (n=0; n<nn; n++) {
    if (!pam)  pm[n] = 0.25*(p00[n] + p01[n] + p10[n] + p11[n]);
    if (p1)  p1[n] = (pm[n] + p00[n] + p10[n])/3;
    if (p2)  p2[n] = (pm[n] + p10[n] + p11[n])/3;
    if (p3)  p3[n] = (pm[n] + p11[n] + p01[n])/3;
    if (p4)  p4[n] = (pm[n] + p01[n] + p00[n])/3;
    }
  if (wa)  FREE(wa);
  return 0;
eX_1: eX_2: eX_3:
  eMSG("invalid parameter! (%s)\n", ePGMs);
eX_10: eX_11: eX_12: eX_13: eX_14:
  eMSG("data not available at (%d,%d,%d)! (%s)\n", i, j, k, ePGMs);
eX_15:
  eMSG("no allocation possible (nn=%d)! (%s)\n", nn, ePGMs);
  }
  
/*====================================================================== GrdMz
*/
int GrdMz( int i, int j, int k, ARYDSC *pap, ARYDSC *pam, float *p )
  {
  dP(GrdMz); 
  int n, nn;
  float *p00, *p01, *p10, *p11, *pm;
  if (!pap || !p)                                eX(1);
  if (!pap->start)                               eX(2);
  if (pap->numdm != 3)                           eX(3);
  nn = pap->elmsz/sizeof(float);
  p00 = AryPtr(pap, i-1, j-1, k);  if (!p00)     eX(10);
  p01 = AryPtr(pap, i-1, j  , k);  if (!p01)     eX(11);
  p10 = AryPtr(pap, i  , j-1, k);  if (!p10)     eX(12);
  p11 = AryPtr(pap, i  , j  , k);  if (!p11)     eX(13);
  if (pam) {
    pm = AryPtr(pam, i, j, k);  if (!pm)         eX(14);
    }
  else  pm = NULL;
  for (n=0; n<nn; n++) {
    if (pm) p[n] = (2*pm[n] + p00[n] + p01[n] + p10[n] + p11[n])/6;
    else    p[n] = (p00[n] + p01[n] + p10[n] + p11[n])/4;
    }
  return 0;
eX_1: eX_2: eX_3:
  eMSG("invalid parameter! (%s)\n", ePGMs);
eX_10: eX_11: eX_12: eX_13: eX_14:
  eMSG("data not available at (%d,%d,%d)! (%s)\n", i, j, k, ePGMs);
  }

/*====================================================================== GrdMv
*/
int GrdMv( int i, int j, int k, ARYDSC *pap, ARYDSC *pam, float *p )
  {
  dP(GrdMv); 
  int n, nn;
  float *p000, *p001, *p010, *p011, *p100, *p101, *p110, *p111, *pm0, *pm1;
  float d00, d01, d10, d11, dm, ds, sumd, w;
  if (!pap || !p)                                   eX(1);
  if (!pap->start)                                  eX(2);
  if (pap->numdm != 3)                              eX(3);
  nn = pap->elmsz/sizeof(float);
  p000 = AryPtr(pap, i-1, j-1, k-1);  if (!p000)    eX(10);
  p001 = AryPtr(pap, i-1, j-1, k  );  if (!p001)    eX(11);
  p010 = AryPtr(pap, i-1, j  , k-1);  if (!p010)    eX(12);
  p011 = AryPtr(pap, i-1, j  , k  );  if (!p011)    eX(13);
  p100 = AryPtr(pap, i  , j-1, k-1);  if (!p100)    eX(14);
  p101 = AryPtr(pap, i  , j-1, k  );  if (!p101)    eX(15);
  p110 = AryPtr(pap, i  , j  , k-1);  if (!p110)    eX(16);
  p111 = AryPtr(pap, i  , j  , k  );  if (!p111)    eX(17);
  if (pam) {
    pm0 = AryPtr(pam, i, j, k-1);  if (!pm0)        eX(18);
    pm1 = AryPtr(pam, i, j, k  );  if (!pm1)        eX(19);
    }
  ds = G.ss[k] - G.ss[k-1];
  if (G.sh > 0) {
    d00 = ds*(G.zh - BP(i-1,j-1))/G.sh;
    d01 = ds*(G.zh - BP(i-1,j  ))/G.sh;
    d10 = ds*(G.zh - BP(i  ,j-1))/G.sh;
    d11 = ds*(G.zh - BP(i  ,j  ))/G.sh;
    dm  = ds*(G.zh - BM(i,j))/G.sh;
    }
  else {
    d00 = ds;
    d01 = ds;
    d10 = ds;
    d11 = ds;
    dm  = ds;
    }
  sumd = d00 + d01 + d10 + d11 + 2*dm;
  for (n=0; n<nn; n++) {
    if (pam) {
      w = 2*dm*(pm0[n]+pm1[n]) + d00*(p000[n]+p001[n]) + d10*(p100[n]+p101[n])
       + d11*(p110[n]+p111[n]) + d01*(p010[n]+p011[n]);
      w /= 2*sumd;
      }
    else {
      w = (dm+2*d00)*(p000[n]+p001[n]) + (dm+2*d10)*(p100[n]+p101[n])
        + (dm+2*d11)*(p110[n]+p111[n]) + (dm+2*d01)*(p010[n]+p011[n]);
      w /= 4*sumd;
      }
    p[n] = w;
    }
  return 0;
eX_1: eX_2: eX_3:
  eMSG("invalid parameter! (%s)\n", ePGMs);
eX_10: eX_11: eX_12: eX_13: eX_14: eX_15: eX_16: eX_17: eX_18: eX_19:
  eMSG("data not available at (%d,%d,%d)! (%s)\n", i, j, k, ePGMs);
  }

/*=================================================================== GrdDivV
*/
float GrdDivV( int i, int j, int k, ARYDSC *pap, ARYDSC *pam )
  {
  dP(GrdDivV);
  float wx0, wx1, wy0, wy1, wz0, wz1, w;
  VECTOR vX, vN1, vN2, vN3, vN4, vZ1, vZ2, vZ3, vZ4;
  if (i<1 || i>G.nx || j<1 || j>G.ny || k<1 || k>G.nz)          eX(1);
  if (!pap)                                                     eX(2);
  if (!pap->start)                                              eX(3);
  AryAssert(pap, sizeof(VECTOR), 3, 0,G.nx, 0,G.ny, 0,G.nz);    eG(4);
  if (pam) {
    if (!pam->start)                                            eG(5);
    AryAssert(pam, sizeof(VECTOR), 3, 1,G.nx, 1,G.ny, 0,G.nz);  eG(6);
    }
  GrdMx(i, j, k, pap, &vX.x);                                   eG(10);
  wx1 = GrdArX(i,j,k)*vX.x;                                     eG(11);
  GrdMx(i-1, j, k, pap, &vX.x);                                 eG(12);
  wx0 = GrdArX(i-1,j,k)*vX.x;                                   eG(13);
  GrdMy(i, j, k, pap, &vX.x);                                   eG(14);
  wy1 = GrdArY(i,j,k)*vX.y;                                     eG(15);
  GrdMy(i, j-1, k, pap, &vX.x);                                 eG(16);
  wy0 = GrdArY(i,j-1,k)*vX.y;                                   eG(17);
  GrdNrm(i, j, k, &vN1, &vN2, &vN3, &vN4);                      eG(18);
  GrdMz4(i, j, k, pap, pam, &vZ1.x, &vZ2.x, &vZ3.x, &vZ4.x);    eG(19);
  wz1 = DotVV(vN1,vZ1) + DotVV(vN2,vZ2) + DotVV(vN3,vZ3) + DotVV(vN4,vZ4);
  GrdNrm(i, j, k-1, &vN1, &vN2, &vN3, &vN4);                    eG(20);
  GrdMz4(i, j, k-1, pap, pam, &vZ1.x, &vZ2.x, &vZ3.x, &vZ4.x);  eG(21);
  wz0 = DotVV(vN1,vZ1) + DotVV(vN2,vZ2) + DotVV(vN3,vZ3) + DotVV(vN4,vZ4);
  w = (wx1 - wx0 + wy1 - wy0 + wz1 - wz0)/GrdVol(i,j,k);        eG(22);
  return w;
eX_1: eX_2: eX_3: eX_4: eX_5: eX_6:
  eMSG("invalid argument! (%s)\n", ePGMs);
eX_10: eX_11: eX_12: eX_13: eX_14: eX_15: eX_16: eX_17: eX_18: eX_19:
eX_20: eX_21: eX_22:
  eMSG("can't calculate surface integral! (%s)\n", ePGMs);
  }
  
/*=================================================================== GrdDivM
*/
VECTOR GrdDivM( int i, int j, int k, ARYDSC *pap, ARYDSC *pam )
  {
  dP(GrdDivM);
  VECTOR vX0, vX1, vY0, vY1, vZ0, vZ1, vW;
  VECTOR vN1, vN2, vN3, vN4, vEx, vEy;
  MATRIX mX, mZ1, mZ2, mZ3, mZ4;
  if (i<1 || i>G.nx || j<1 || j>G.ny || k<1 || k>G.nz)          eX(1);
  if (!pap)                                                     eX(2);
  if (!pap->start)                                              eX(3);
  AryAssert(pap, sizeof(MATRIX), 3, 0,G.nx, 0,G.ny, 0,G.nz);    eG(4);
  if (pam) {
    if (!pam->start)                                            eG(5);
    AryAssert(pam, sizeof(MATRIX), 3, 1,G.nx, 1,G.ny, 0,G.nz);  eG(6);
    }
  vEx = DefV(1, 0, 0);
  vEy = DefV(0, 1, 0);
  GrdMx(i, j, k, pap, &mX.xx);                                  eG(10);
  vX1 = MulSV(GrdArX(i,j,k),DotVM(vEx, mX));                    eG(11);
  GrdMx(i-1, j, k, pap, &mX.xx);                                eG(12);
  vX0 = MulSV(GrdArX(i-1,j,k),DotVM(vEx,mX));                   eG(13);
  GrdMy(i, j, k, pap, &mX.xx);                                  eG(14);
  vY1 = MulSV(GrdArY(i,j,k),DotVM(vEy,mX));                     eG(15);
  GrdMy(i, j-1, k, pap, &mX.xx);                                eG(16);
  vY0 = MulSV(GrdArY(i,j-1,k),DotVM(vEy,mX));                   eG(17);
  GrdNrm(i, j, k, &vN1, &vN2, &vN3, &vN4);                      eG(18);
  GrdMz4(i, j, k, pap, pam, &mZ1.xx, &mZ2.xx, &mZ3.xx, &mZ4.xx);    eG(19);
  vZ1 = DotVM(vN1,mZ1);
  vZ1 = AddVV(vZ1, DotVM(vN2,mZ2));
  vZ1 = AddVV(vZ1, DotVM(vN3,mZ3));
  vZ1 = AddVV(vZ1, DotVM(vN4,mZ4));
  GrdNrm(i, j, k-1, &vN1, &vN2, &vN3, &vN4);                    eG(20);
  GrdMz4(i, j, k-1, pap, pam, &mZ1.xx, &mZ2.xx, &mZ3.xx, &mZ4.xx);  eG(21);
  vZ0 = DotVM(vN1,mZ1);
  vZ0 = AddVV(vZ0, DotVM(vN2,mZ2));
  vZ0 = AddVV(vZ0, DotVM(vN3,mZ3));
  vZ0 = AddVV(vZ0, DotVM(vN4,mZ4));
  vW = SubVV(vX1, vX0);
  vW = AddVV(vW, SubVV(vY1, vY0));
  vW = AddVV(vW, SubVV(vZ1, vZ0));
  vW = MulSV(1/GrdVol(i,j,k), vW);                              eG(22);
  return vW;
eX_1: eX_2: eX_3: eX_4: eX_5: eX_6:
  vMsg("invalid argument! (%s)\n", ePGMs);
  return vEx;
eX_10: eX_11: eX_12: eX_13: eX_14: eX_15: eX_16: eX_17: eX_18: eX_19:
eX_20: eX_21: eX_22:
  vMsg("can't calculate surface integral! (%s)\n", ePGMs);
  return vEx;
  }
  
  
#ifdef  MAIN  /*###########################################################*/

#include "IBJdmn.h"

#define  VCP(i,j,k)  (*(VECTOR*)AryPtrX(&VcpDsc,(i),(j),(k)))
#define  VCM(i,j,k)  (*(VECTOR*)AryPtrX(&VcmDsc,(i),(j),(k)))
#define  MTP(i,j,k)  (*(MATRIX*)AryPtrX(&MtpDsc,(i),(j),(k)))
#define  MTM(i,j,k)  (*(MATRIX*)AryPtrX(&MtmDsc,(i),(j),(k)))
#define  DVS(i,j,k)  (*(float*)AryPtrX(&DvsDsc,(i),(j),(k)))
#define  DVV(i,j,k)  (*(VECTOR*)AryPtrX(&DvvDsc,(i),(j),(k)))

ARYDSC VcpDsc, VcmDsc, DvsDsc, BpDsc, MtpDsc, MtmDsc, DvvDsc;
float X0, Y0, X1, Y1, Dd, Zh, Sh, Ss[100];
int Nx, Ny, Nz;
char Path[256] = ".";

int main( int argc, char *argv[] )
  {
  dP(TstGrd);
  int i, j, k;
  char fname[256];
  TXTSTR usrhdr = { NULL, 0 };
  TXTSTR syshdr = { NULL, 0 };
  VECTOR vR;
  MATRIX mR;
  if (argc > 1)  strcpy(Path, argv[1]);
  sprintf(fname, "%s%c%s", Path, PATHSEP, "bp.dmd");
  DmnRead(fname, &usrhdr, &syshdr, &BpDsc);                       eG(1);
  printf("surface definition\n");
  Dd = 200;
  Nx = BpDsc.bound[0].hgh;
  Ny = BpDsc.bound[1].hgh;
  Nz = 5;
  Ss[0]=0;  Ss[1]=10;  Ss[2]=20;  Ss[3]=50;  Ss[4]=100;  Ss[5]=110;
  Sh = 500; 
  Zh = 500;
  GrdDefK(0, X0, Y0, X1, Y1, Dd, Nz, Ss, Sh, Zh, &BpDsc, NULL);     eG(2);
  printf("grid defined\n");
  GrdSet(0);
  AryCreate(&MtpDsc, sizeof(MATRIX), 3, 0, Nx, 0, Ny, 0, Nz);       eG(3);
  AryCreate(&MtmDsc, sizeof(MATRIX), 3, 1, Nx, 1, Ny, 0, Nz);       eG(13);
  AryCreate(&DvvDsc, sizeof(VECTOR), 3, 1, Nx, 1, Ny, 1, Nz);       eG(4);
  printf("arrays created\n");
  for (i=0; i<=Nx; i++)
    for (j=0; j<=Ny; j++)
      for (k=0; k<=Nz; k++) {
        vR = GrdVp(i, j, k);                                        eG(5);
        mR = DefM(0,0,0,0,0,0,0,0,0);
        mR.zy = vR.z;
        MTP(i,j,k) = mR;
        if (i>0 && j>0) {
          vR = GrdVm(i, j, k);                                      eG(15);
          mR = DefM(0,0,0,0,0,0,0,0,0);
          mR.zy = vR.z;
          MTM(i,j,k) = mR;
          }
        }
  printf("vector field defined\n");
  for (i=1; i<=Nx; i++)
    for (j=1; j<=Ny; j++)
      for (k=1; k<=Nz; k++) {
        DVV(i,j,k) = GrdDivM(i, j, k, &MtpDsc, &MtmDsc);            eG(6);
        }
  TxtCpy(&usrhdr, "mo\ttext\n");
  TxtCat(&usrhdr, "fo\t\~%13.5e\n");
  sprintf(fname, "%s%c%s", Path, PATHSEP, "vdiv");
  DmnWrite(fname, &usrhdr, &DvvDsc);                                   eG(7);
  printf("done\n");
  return 0;
eX_1: eX_2: eX_3: eX_4: eX_5: eX_6: eX_7: eX_8:
eX_13: eX_15:
  eMSG("error! (%s)\n", ePGMs);
  }
#endif /*#################################################################*/
