//========================================================== IBJp3d.c
//
// 3d functions for meteorology
// ============================
//
// Copyright (C) Dr. Lutz Janicke, Dunum, Germany, 1999
// email: lj@janicke.de
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// history:
//
// last change: 98-08-15  lj
//
//==================================================================

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <ctype.h>

#include "IBJprm.h"
#include "IBJmsg.h"
#include "IBJary.h"
#include "IBJrnd.h"
#include "IBJdmn.h"
#include "IBJmth.h"
#include "IBJgrd.h"
#include "IBJprt.h"
#include "IBJtbl.h"
#include "IBJp3d.h"

#define  VERSION  1.01

#define  PERX  (PrtMode & PRT_PEX)
#define  PERY  (PrtMode & PRT_PEY)
#define  MET3  (PrtMode & PRT_3DM)
#define  G     (*pGrd)

static char *eMODn = "IBJp3d";

/*==========================================================================*/

#define  BP(i,j)  (*(float*)AryPtrX(&(G.bpdsc),(i),(j)))
#define  BM(i,j)  (*(float*)AryPtrX(&(G.bmdsc),(i),(j)))

ARYDSC  VxDsc;
#define VX3(a,b,c)  (*(float*)AryPtrX(&VxDsc,(a),(b),(c)))
ARYDSC  VyDsc;
#define VY3(a,b,c)  (*(float*)AryPtrX(&VyDsc,(a),(b),(c)))
ARYDSC  VsDsc;
#define VS3(a,b,c)  (*(float*)AryPtrX(&VsDsc,(a),(b),(c)))
ARYDSC  SuDsc;
#define SU3(a,b,c)  (*(float*)AryPtrX(&SuDsc,(a),(b),(c)))
ARYDSC  SvDsc;
#define SV3(a,b,c)  (*(float*)AryPtrX(&SvDsc,(a),(b),(c)))
ARYDSC  SwDsc;
#define SW3(a,b,c)  (*(float*)AryPtrX(&SwDsc,(a),(b),(c)))
ARYDSC  ScDsc;
#define SC3(a,b,c)  (*(float*)AryPtrX(&ScDsc,(a),(b),(c)))
ARYDSC  KuDsc;
#define KU3(a,b,c)  (*(float*)AryPtrX(&KuDsc,(a),(b),(c)))
ARYDSC  KvDsc;
#define KV3(a,b,c)  (*(float*)AryPtrX(&KvDsc,(a),(b),(c)))
ARYDSC  KwDsc;
#define KW3(a,b,c)  (*(float*)AryPtrX(&KwDsc,(a),(b),(c)))
ARYDSC  TtDsc;
#define TT3(a,b,c)  (*(float*)AryPtrX(&TtDsc,(a),(b),(c)))
ARYDSC  LiDsc;
#define LI3(a,b,c)  (*(float*)AryPtrX(&LiDsc,(a),(b),(c)))
ARYDSC  VdDsc;
#define VD2(a,b)    (*(float*)AryPtrX(&VdDsc,(a),(b)))
ARYDSC  PdDsc;
#define PD2(a,b)    (*(float*)AryPtrX(&PdDsc,(a),(b)))

#define WW3(a,b,c)  (*(VECTOR*)AryPtrX(&WwDsc,(a),(b),(c)))
#define SS3(a,b,c)  (*(MATRIX*)AryPtrX(&SsDsc,(a),(b),(c)))
#define LL3(a,b,c)  (*(MATRIX*)AryPtrX(&LlDsc,(a),(b),(c)))
#define KK3(a,b,c)  (*(MATRIX*)AryPtrX(&KkDsc,(a),(b),(c)))
#define PP3(a,b,c)  (*(MATRIX*)AryPtrX(&PpDsc,(a),(b),(c)))
#define EE3(a,b,c)  (*(MATRIX*)AryPtrX(&EeDsc,(a),(b),(c)))

static ARYDSC TsDsc;
#define TS3(a,b,c)  (*(MATRIX*)AryPtrX(&TsDsc,(a),(b),(c)))
static ARYDSC FsDsc;
#define FS3(a,b,c)  (*(MATRIX*)AryPtrX(&FsDsc,(a),(b),(c)))
static ARYDSC GsDsc;
#define GS3(a,b,c)  (*(MATRIX*)AryPtrX(&GsDsc,(a),(b),(c)))

static float Ta, Aa, Vd, Wd, Vs;

/*---------------------------------------------------------------------------*/

static void prnmat(MATRIX m)
  {
  fprintf(ChkFile, "%12.5e  %12.5e  %12.5e\n", m.xx, m.xy, m.xz);
  fprintf(ChkFile, "%12.5e  %12.5e  %12.5e\n", m.yx, m.yy, m.yz);
  fprintf(ChkFile, "%12.5e  %12.5e  %12.5e\n", m.zx, m.zy, m.zz);
  }
  
static int calculate_Vs( void )
  {
  dP(clcVs);
  int i, j, k;
  float fx0, fx1, fy0, fy1, fs0, fs1, s2h, hm;
  if (VsDsc.start)  return 0;
  AryCreate(&VsDsc, sizeof(float),  3, 1, G.nx, 1, G.ny, 0, G.nz);    eG(1);
  for (i=1; i<=G.nx; i++)
    for (j=1; j<=G.ny; j++) {
      if (G.sh <= 0)  s2h = 1;
      else {
        hm = (BP(i-1,j-1) + BP(i,j-1) + BP(i-1,j) + BP(i,j) + 2*BM(i,j))/6;
        s2h = (G.zh - hm)/G.sh;
        }
      fs0 = 0;
      VS3(i,j,0) = 0;
      for (k=1; k<=G.nz; k++) {
        fx1 = GrdArX(i,j,k)*VX3(i,j,k);
        fx0 = GrdArX(i-1,j,k)*VX3(i-1,j,k);
        fy1 = GrdArY(i,j,k)*VY3(i,j,k);
        fy0 = GrdArY(i,j-1,k)*VY3(i,j-1,k);
        fs1 = fs0 - (fx1-fx0+fy1-fy0);
        VS3(i,j,k) = fs1/(s2h*G.dd*G.dd);
        fs0 = fs1;
        }
      }
  return 0;
eX_1:
  eMSG("can't allocate (%d,%d,%d)! (%s)\n", G.nx, G.ny, G.nz, ePGMs);
  }

static int allocate_all( void )
  {
  dP(allocate_all);
  AryCreate(&TtDsc, sizeof(float),  3, 0, G.nx, 0, G.ny, 0, G.nz);    eG(11);
  AryCreate(&WwDsc, sizeof(VECTOR), 3, 1, G.nx, 1, G.ny, 1, G.nz);    eG(15);
  AryCreate(&SsDsc, sizeof(MATRIX), 3, 0, G.nx, 0, G.ny, 0, G.nz);    eG(16);
  AryCreate(&LlDsc, sizeof(MATRIX), 3, 0, G.nx, 0, G.ny, 0, G.nz);    eG(17);
  AryCreate(&KkDsc, sizeof(MATRIX), 3, 0, G.nx, 0, G.ny, 0, G.nz);    eG(18);
  AryCreate(&PpDsc, sizeof(MATRIX), 3, 0, G.nx, 0, G.ny, 0, G.nz);    eG(19);
  AryCreate(&EeDsc, sizeof(MATRIX), 3, 0, G.nx, 0, G.ny, 0, G.nz);    eG(20);
  if (Wd >= 0) {
    AryCreate(&PdDsc, sizeof(float),  2, 1, G.nx, 1, G.ny);           eG(21);
    }
  return 0;
eX_11: eX_15: eX_16: eX_17: eX_18: eX_19: eX_20:
eX_21:
  eMSG("can't allocate (%d,%d,%d)! (%s)\n", G.nx, G.ny, G.nz, ePGMs);
  }

/*================================================================= PrtClc3dm
*/
#define  VALIDX(i,j,k)  ((j>0) && (j<=ny) && (k>0) && (k<=nz) &&\
 (((i>0) && (VS3((i),(j),(k)-1)>-99)) || ((i<nx) && (VS3((i)+1,(j),(k)-1)>-99))))
#define  VALIDY(i,j,k)  ((i>0) && (i<=nx) && (k>0) && (k<=nz) &&\
 (((j>0) && (VS3((i),(j),(k)-1)>-99)) || ((j<ny) && (VS3((i),(j)+1,(k)-1)>-99))))
#define  VALIDZ(i,j,k)  ((i>0) && (i<=nx) && (j>0) && (j<=ny) &&\
 (((k>0) && (VS3((i),(j),(k)-1)>-99)) || ((k<nz) && (VS3((i),(j),(k))>-99))))

int PrtClc3dm( void )	// calculate 3d model fields
  {
  dP(PrtClc3dm);
  int i, j, k, nx, ny, nz;
  int su_defined, sc_defined, ku_defined, vd_defined;
  float tau, su, sv, sw, sc, ku, kv, kw, tu, tv, tw;
  float sm, vm, vr, s2zfac, d, g, gp, gm, vx, vy;
  float s0, w2, wp, ga, fs, vd;
  VECTOR div1, div2;
  MATRIX sigma, psi, phi, omega, rot, trot, kappa;
  MATRIX ident, denom, fak1, fak2;
  float tmax;
  tmax = 1.e6;
  Ta = TA;
  Aa = AA;
  Vd = VD;
  Wd = PD;
  if (Wd<0 && !PdDsc.start)				eX(5);
  Vs = VS; 
  nx = G.nx;
  ny = G.ny; 
  nz = G.nz;
  if (!TtDsc.start) {
    allocate_all();         			eG(6);
    }
  su_defined = (SuDsc.start != NULL);
  sc_defined = (ScDsc.start != NULL);
  ku_defined = (KuDsc.start != NULL);
  vd_defined = (VdDsc.start != NULL);
  if (su_defined) {
    AryAssert(&SuDsc, sizeof(float), 3, 0,nx, 0,ny, 0,nz);    eG(10);
    }
  if (sc_defined) {
    AryAssert(&ScDsc, sizeof(float), 3, 0,nx, 0,ny, 0,nz);    eG(11);
    }
  if (ku_defined) {
    AryAssert(&KuDsc, sizeof(float), 3, 0,nx, 0,ny, 0,nz);    eG(12);
    }
  if (vd_defined) {
    AryAssert(&VdDsc, sizeof(float), 2, 1,nx, 1,ny);          eG(13);
    }
  AryAssert(&SvDsc, sizeof(float), 3, 0,nx, 0,ny, 0,nz);      eG(14);
  AryAssert(&SwDsc, sizeof(float), 3, 0,nx, 0,ny, 0,nz);      eG(15);
  AryAssert(&KvDsc, sizeof(float), 3, 0,nx, 0,ny, 0,nz);      eG(16);
  AryAssert(&KwDsc, sizeof(float), 3, 0,nx, 0,ny, 0,nz);      eG(17);
  AryAssert(&VxDsc, sizeof(float), 3, 0,nx, 1,ny, 1,nz);      eG(18);
  AryAssert(&VyDsc, sizeof(float), 3, 1,nx, 0,ny, 1,nz);      eG(19);
  if (VsDsc.start) {
    AryAssert(&VsDsc, sizeof(float), 3, 1,nx, 1,ny, 0,nz);    eG(20);
    }
  else {
    calculate_Vs();                                           eG(20);
    }
  AryAssert(&PdDsc, sizeof(float),  2, 1,nx, 1,ny);           eG(21);
      
  AryCreate(&TsDsc, sizeof(MATRIX), 3, 0,nx, 0,ny, 0,nz);     eG(30);
  AryCreate(&FsDsc, sizeof(MATRIX), 3, 0,nx, 0,ny, 0,nz);     eG(31);
  AryCreate(&GsDsc, sizeof(MATRIX), 3, 0,nx, 0,ny, 0,nz);     eG(32);

  ident = DefM(1,0,0, 0,1,0, 0,0,1);
  for (i=0; i<=nx; i++)
    for (j=0; j<=ny; j++) {
      s2zfac = GrdS2Zij(i, j);
      for (k=0; k<=nz; k++) {        
        gm = (k > 0) ? G.ss[k] - G.ss[k-1] : 0;
        gp = (k < nz) ? G.ss[k+1] - G.ss[k] : 0;
        g = 0;
        vx = 0;
        if (VALIDX(i,j  ,k  ))  { vx += gm*VX3(i,j  ,k  );  g += gm; }
        if (VALIDX(i,j  ,k+1))  { vx += gp*VX3(i,j  ,k+1);  g += gp; }
        if (VALIDX(i,j+1,k  ))  { vx += gm*VX3(i,j+1,k  );  g += gm; }
        if (VALIDX(i,j+1,k+1))  { vx += gp*VX3(i,j+1,k+1);  g += gp; }
		if (g <= 0)  continue;
        vx /= g;
        g = 0;
        vy = 0;
        if (VALIDY(i  ,j,k  ))  { vy += gm*VY3(i  ,j,k  );  g += gm; }
        if (VALIDY(i  ,j,k+1))  { vy += gp*VY3(i  ,j,k+1);  g += gp; }
        if (VALIDY(i+1,j,k  ))  { vy += gm*VY3(i+1,j,k  );  g += gm; }
        if (VALIDY(i+1,j,k+1))  { vy += gp*VY3(i+1,j,k+1);  g += gp; }
		if (g <= 0)  continue;
        vy /= g;

        sv = SV3(i,j,k);
        sw = SW3(i,j,k);
if (PrtCheck & 0x1000) fprintf(ChkFile,
"(%d,%d,%d):  sv=%12.3e  sw=%12.3e\n", i, j, k, sv, sw);
        kv = KV3(i,j,k);
        kw = KW3(i,j,k);
        if (su_defined)  su = SU3(i,j,k);
        else             su = sv;
        if (sc_defined)  sc = SC3(i,j,k);
        else             sc = 0;
        if (ku_defined)  ku = KU3(i,j,k);
        else             ku = kv;
        if (su > 0)  tu = ku/(su*su);
        else         tu = tmax;
        if (sv > 0)  tv = kv/(sv*sv);
        else         tv = tmax;
        if (sw > 0)  tw = kw/(sw*sw);
        else         tw = tmax;
		
        if (Ta > 0)  tau = Ta;          /* absoluter Zeitschritt    */
        else {
          tau = tw;
          if (tv < tau)  tau = tv;
          if (tu < tau)  tau = tu;
          if (Aa > 0)  tau *= Aa;       /* relativer Zeitschritt    */
          else {                        /* angepater Zeitschritt ... */
            tau *= 2;
            d = 0.5*G.dd;
            sm = MAX(su, sv);
            vm = 0;
            if (VALIDX(i,j  ,k  ))  vm = MAX(vm, fabs(VX3(i,j  ,k  )));
            if (VALIDX(i,j  ,k+1))  vm = MAX(vm, fabs(VX3(i,j  ,k+1)));
            if (VALIDX(i,j+1,k  ))  vm = MAX(vm, fabs(VX3(i,j+1,k  )));
            if (VALIDX(i,j+1,k+1))  vm = MAX(vm, fabs(VX3(i,j+1,k+1)));

            if (VALIDY(i  ,j,k  ))  vm = MAX(vm, fabs(VY3(i  ,j,k  )));
            if (VALIDY(i  ,j,k+1))  vm = MAX(vm, fabs(VY3(i  ,j,k+1)));
            if (VALIDY(i+1,j,k  ))  vm = MAX(vm, fabs(VY3(i+1,j,k  )));
            if (VALIDY(i+1,j,k+1))  vm = MAX(vm, fabs(VY3(i+1,j,k+1)));

            vr = vm + sm;
            if (tau*vr > d)  tau = d/vr;
if (PrtCheck & 0x2000)  fprintf(ChkFile,
"A (%d,%d,%d)  d=%1.3f  sm=%1.5f  vm=%1.5f  vr=%1.5f  tau=%1.5f\n", 
i, j, k, d, sm, vm, vr, tau);
            sm = sw/s2zfac;
            if (k > 0) {
              d = 0.5*(G.ss[k] - G.ss[k-1]);
              vm = 0;

              if (VALIDZ(i  ,j  ,k))  vm = MIN(vm, VS3(i  ,j  ,k));
              if (VALIDZ(i  ,j+1,k))  vm = MIN(vm, VS3(i  ,j+1,k));
              if (VALIDZ(i+1,j  ,k))  vm = MIN(vm, VS3(i+1,j  ,k));
              if (VALIDZ(i+1,j+1,k))  vm = MIN(vm, VS3(i+1,j+1,k));

              vr = -vm + sm;
              if (tau*vr > d)  tau = d/vr;
if (PrtCheck & 0x2000)  fprintf(ChkFile,
"B (%d,%d,%d)  d=%1.3f  sm=%1.5f  vm=%1.5f  vr=%1.5f  tau=%1.5f\n", 
i, j, k, d, sm, vm, vr, tau);
              }
            if (k < nz) {
              d = 0.5*(G.ss[k+1] - G.ss[k]);
              vm = 0;
			  
              if (VALIDZ(i  ,j  ,k))   vm = MAX(vm, VS3(i  ,j  ,k));
              if (VALIDZ(i  ,j+1,k))   vm = MAX(vm, VS3(i  ,j+1,k));
              if (VALIDZ(i+1,j  ,k))   vm = MAX(vm, VS3(i+1,j  ,k));
              if (VALIDZ(i+1,j+1,k))   vm = MAX(vm, VS3(i+1,j+1,k));

              vr = vm + sm;
              if (tau*vr > d)  tau = d/vr;
if (PrtCheck & 0x2000)  fprintf(ChkFile,
"C (%d,%d,%d)  d=%1.3f  sm=%1.5f  vm=%1.5f  vr=%1.5f  tau=%1.5f\n", 
i, j, k, d, sm, vm, vr, tau);
              }
            }
          }
        TT3(i,j,k) = tau;
        
        rot   = DefR2(vx, vy);
        trot  = TrnM(rot);
        sigma = DefM(su*su, 0, -sc*sc, 0, sv*sv, 0, -sc*sc, 0, sw*sw);
        sigma = DotMM(rot, DotMM(sigma, trot));
        kappa = DefM(ku, 0, 0, 0, kv, 0, 0, 0, kw);
        kappa = DotMM(rot, DotMM(kappa, trot));
        phi   = DotMM(sigma, InvM(kappa));
        denom = InvM (AddMM(ident, MulSM(0.5*tau, phi)));
        psi   = DotMM(SubMM(ident, MulSM(0.5*tau, phi)), denom);
if ((PrtCheck & 0x4000) && (i==1) && (j==1) && (k==1)) {
fprintf(ChkFile, "tau=%12.5e\n", tau);
fprintf(ChkFile, "sigma:\n");  prnmat(sigma);
fprintf(ChkFile, "kappa:\n");  prnmat(kappa);
fprintf(ChkFile, "phi:\n");  prnmat(phi);
fprintf(ChkFile, "psi:\n");  prnmat(psi);
}
        omega = SubMM(sigma, DotMM(psi, DotMM(sigma, TrnM(psi))));
        SS3(i, j, k) = sigma;
        KK3(i, j, k) = kappa;
        PP3(i, j, k) = psi;
        EE3(i, j, k) = ChlM(sigma);
        LL3(i, j, k) = ChlM(omega);
        TS3(i,j,k) = MulSM(tau, sigma);
        FS3(i,j,k) = MulSM(0.5, SubMM(ident, psi));
        GS3(i,j,k) = MulSM(tau, psi);   
        }
      }
      
  for (i=1; i<=nx; i++)
    for (j=1; j<=ny; j++)
      for (k=1; k<=nz; k++) {
		if (VS3(i,j,k-1) <= -99)  continue;
        GrdMv(i, j, k, &FsDsc, NULL, &fak1.xx);     eG(40);
        GrdMv(i, j, k, &GsDsc, NULL, &fak2.xx);     eG(41);
        div1 = GrdDivM(i, j, k, &TsDsc, NULL);      eG(42);
        div2 = GrdDivM(i, j, k, &SsDsc, NULL);      eG(43);
        WW3(i,j,k) = AddVV(DotMV(fak1,div1), DotMV(fak2, div2));
        }
        
  AryFree(&TsDsc);
  AryFree(&FsDsc);
  AryFree(&GsDsc);
  
  if (Wd < 0)  return 0;
            
  w2 = sqrt(2.0);
  wp = sqrt(Pi);
  for (i=1; i<=G.nx; i++)
    for (j=1; j<=G.ny; j++) {
	  if (VS3(i,j,0) <= -99)  continue;
      if ((Wd<=0) && (Vd>0 || vd_defined)) {
        if (Vd > 0)  vd = Vd;
        else         vd = VD2(i,j);
        s0 = SW3(i,j,0);
        ga = Vs/(w2*s0);
        fs = wp*ga + exp(-ga*ga)/(2 - Erfc(ga));
        Wd = 2*wp*vd/(w2*s0*fs + wp*vd);
        if (Wd > 1) {
          vMsg("Wd(%d,%d)=%1.2f !\n", i, j, Wd);
          Wd = 1;
          }
        }
      PD2(i,j) = Wd;
      }
  return 0;
eX_5:
  eMSG("missing wd-field! (%s)\n", ePGMs);
eX_6:
eX_30: eX_31: eX_32:
  eMSG("no allocation possible! (%s)\n", ePGMs);
eX_10: eX_11: eX_12: eX_13: eX_14: eX_15: eX_16: eX_17: eX_18: eX_19: 
eX_20: eX_21:
  eMSG("invalid structure! (%s)\n", ePGMs);
eX_40: eX_41: eX_42: eX_43:
  eMSG("can't evaluate values at (%d,%d,%d)! (%s)\n",  i, j, k, ePGMs);
  }
#undef  VALIDX  
#undef  VALIDY  
#undef  VALIDZ  
/*==============================================================================*/





