// Copyright Dr. Lutz Janicke, Dunum, Germany, 1998-2000

// This is the header IBJtab.h to be used by the program "IBJparticle".
// It defines all parameters whose values can be given by the user.
// It has to be included by all modules that use these values.

// Each parameter is defined in a record of the struct PrmTab.
// The name of the record is the same as the name of the input
// parameter (small letters). The record contains the name, a
// short description, input and output format and the default
// value as a string. For arrays also the number of elements or
// a reference to a parameter is given, that represents this
// number. For details see the header IBJprm.h.

#ifndef  IBJPRM_INCLUDE
    #include "IBJprm.h"  // The typedef of PRMTABREC is needed
#endif

#ifndef  IBJPRM_SOURCE
    extern  // Only IBJPRM saves the data of this structure
#endif
struct { PRMTABREC  // The names of the input parameters:
 _c, _d, mc, mx, my, mz, nx, ny, nz, op, sd, ti, _g, a0, b0, cc, da, dc, dd, sh,
 x0, x1, y0, y1, zh, zz, _p, aa, ku, kv, kw, pd, qp, rp, rw, sc, su, sv, sw, ta,
 tt, tu, tv, tw, vd, vs, vx, vy, _q, aq, bq, cq, eq, hq, pq, sq, uu, ud, ut, xq,
 yq, _r, fn, rc, _w, aw, fa, fi, fo, id, mo, se, wc, _z, dt, _t, _l, _n, _e, end_of_table; } PrmTab
#ifdef  IBJPRM_SOURCE
//      The initial values are defined in IBJPRM
 = {
 { "*c",  "%s",  " \"%s\"",    "CLR",    "0", 255,  0, 0x005, 0, "Clear [all]" },
 { "*d",  "%s",  " \"%s\"",   "DIMS",    "0", 255,  0, 0x005, 0, "Dimensionierung von Feldern und Rechengebiet" },
 { "mc",  "%d",    " %10d",   "4000",    "0",   4,  0, 0x002, 0, "Maximale Anzahl von Zeichen pro Zeile" },
 { "mx",  "%d",    " %10d",      "0",    "0",   4,  0, 0x002, 0, "Auszhlgitter: Anzahl der Intervalle in x-Richtung" },
 { "my",  "%d",    " %10d",      "0",    "0",   4,  0, 0x002, 0, "Auszhlgitter: Anzahl der Intervalle in y-Richtung" },
 { "mz",  "%d",    " %10d",      "0",    "0",   4,  0, 0x002, 0, "Auszhlgitter: Anzahl der Intervalle in z-Richtung" },
 { "nx",  "%d",    " %10d",      "0",    "0",   4,  0, 0x002, 0, "Rechengitter: Anzahl der Intervalle in x-Richtung" },
 { "ny",  "%d",    " %10d",      "0",    "0",   4,  0, 0x002, 0, "Rechengitter: Anzahl der Intervalle in y-Richtung" },
 { "nz",  "%d",    " %10d",      "0",    "0",   4,  0, 0x002, 0, "Rechengitter: Anzahl der Intervalle in z-Richtung" },
 { "op",  "%s",  " \"%s\"",       "",    "0", 255,  0, 0x005, 0, "Options: perx pery met3 velcor" },
 { "sd",  "%d",    " %10d",  "11111",    "0",   4,  0, 0x002, 0, "Startzahl fr Zufallszahlen" },
 { "ti",  "%s",  " \"%s\"",   "TEST",    "0", 255,  0, 0x005, 0, "Titel" },
 { "*g",  "%s",  " \"%s\"",   "GRID",    "0", 255,  0, 0x005, 0, "Gitter-Festlegung" },
 { "a0",  "%f",  " %10.0f",      "0",    "0",   4,  0, 0x003, 0, "Auszhl-Gitter: Kleinste x-Koordinate (m)" },
 { "b0",  "%f",  " %10.0f",      "0",    "0",   4,  0, 0x003, 0, "Auszhl-Gitter: Kleinste y-Koordinate (m)" },
 { "cc",  "%f",   " %6.1f",      "0", "mz+1",   4, -1, 0x103, 0, "Auszhl-Gitter: Vertikale Gitterpunkte (m)" },
 { "da",  "%f",  " %10.0f",      "0",    "0",   4,  0, 0x003, 0, "Auszhl-Gitter: Horizontale Maschenweite (m)" },
 { "dc",  "%f",  " %10.0f",      "0",    "0",   4,  0, 0x003, 0, "Auszhl-Gitter: Vertikale Maschenweite (m)" },
 { "dd",  "%f",   " %4.1f",      "0",    "0",   4,  0, 0x003, 0, "Horizontale Maschenweite" },
 { "sh",  "%f",  " %10.0f",      "0",    "0",   4,  0, 0x003, 0, "Skalierungswert s-Dach (m)" },
 { "x0",  "%f",  " %10.0f",      "0",    "0",   4,  0, 0x003, 0, "Rechen-Gitter: Kleinste x-Koordinate (m)" },
 { "x1",  "%f",  " %10.4f",      "0",    "0",   4,  0, 0x003, 0, "Rechen-Gitter: Grte x-Koordinate (m)" },
 { "y0",  "%f",  " %10.0f",      "0",    "0",   4,  0, 0x003, 0, "Rechen-Gitter: Kleinste y-Koordinate (m)" },
 { "y1",  "%f",  " %10.2f",      "0",    "0",   4,  0, 0x003, 0, "Rechen-Gitter: Grte y-Koordinate (m)" },
 { "zh",  "%f",  " %10.0f",      "0",    "0",   4,  0, 0x003, 0, "Skalierungswert z-Dach (m)" },
 { "zz",  "%f",   " %6.1f",      "0", "nz+1",   4, -1, 0x103, 0, "Vertikales Netz" },
 { "*p",  "%s",  " \"%s\"",    "PRM",    "0", 255,  0, 0x005, 0, "Parameter der Ausbreitungsrechnung" },
 { "aa",  "%f",  " %10.2f",    "0.0",    "0",   4,  0, 0x003, 0, "Relativer Zeitschritt" },
 { "ku",  "%f",   " %6.1f",     "10", "nz+1",   4, -1, 0x103, 0, "Diffusions-Koeffizient Ku (m2/s)" },
 { "kv",  "%f",   " %6.1f",     "10", "nz+1",   4, -1, 0x103, 0, "Diffusions-Koeffizient Kv (m2/s)" },
 { "kw",  "%f",   " %6.1f",     "10", "nz+1",   4, -1, 0x103, 0, "Diffusions-Koeffizient Kw (m2/s)" },
 { "pd",  "%f",  " %10.3f",    "0.0",    "0",   4,  0, 0x003, 0, "Depositions-Wahrscheinlichkeit" },
 { "qp",  "%f",  " %10.5f",   "0.01",    "0",   4,  0, 0x003, 0, "Kleinste relative Masse eines Partikels" },
 { "rp",  "%f",  " %10.2f",    "1.0",    "0",   4,  0, 0x003, 0, "Emissionsrate der Teilchen (1/s)" },
 { "rw",  "%f",  " %10.5f",    "0.0",    "0",   4,  0, 0x003, 0, "Auswaschrate (1/s)" },
 { "sc",  "%f",   " %6.2f",      "0", "nz+1",   4, -1, 0x103, 0, "Korrelation sqrt(-<u'w'>) (m/s)" },
 { "su",  "%f",   " %6.2f",    "0.5", "nz+1",   4, -1, 0x103, 0, "Geschwindigkeitsstreuung Sigma-u (m/s)" },
 { "sv",  "%f",   " %6.2f",    "0.5", "nz+1",   4, -1, 0x103, 0, "Geschwindigkeitsstreuung Sigma-v (m/s)" },
 { "sw",  "%f",   " %6.2f",    "0.5", "nz+1",   4, -1, 0x103, 0, "Geschwindigkeitsstreuung Sigma-w (m/s)" },
 { "ta",  "%f",  " %10.1f",    "1.0",    "0",   4,  0, 0x003, 0, "Absoluter Zeitschritt (s)" },
 { "tt",  "%f",   " %6.1f",      "0", "nz+1",   4, -1, 0x103, 0, "Zeitschritt Tau (s)" },
 { "tu",  "%f",   " %6.1f",      "0", "nz+1",   4, -1, 0x103, 0, "Lagrange-Korrelationszeit Tu (s)" },
 { "tv",  "%f",   " %6.1f",      "0", "nz+1",   4, -1, 0x103, 0, "Lagrange-Korrelationszeit Tv (s)" },
 { "tw",  "%f",   " %6.1f",      "0", "nz+1",   4, -1, 0x103, 0, "Lagrange-Korrelationszeit Tw (s)" },
 { "vd",  "%f",  " %10.3f",    "0.0",    "0",   4,  0, 0x003, 0, "Depositions-Geschwindigkeit (m/s)" },
 { "vs",  "%f",  " %10.3f",    "0.0",    "0",   4,  0, 0x003, 0, "Sedimentations-Geschwindigkeit (m/s)" },
 { "vx",  "%f",   " %6.2f",    "1.0", "nz+1",   4, -1, 0x103, 0, "Geschwindigkeit in x-Richtung (m/s)" },
 { "vy",  "%f",   " %6.2f",    "0.0", "nz+1",   4, -1, 0x103, 0, "Geschwindigkeit in y-Richtung (m/s)" },
 { "*q",  "%s",  " \"%s\"",      "0",    "0", 255,  0, 0x005, 0, "Quellen: Lage, berhhung und Quellstrken" },
 { "aq",  "%f",  " %10.1f",      "0",    "0",   4,  0, 0x003, 0, "Horizontale Ausdehnung der Quelle (m)" },
 { "bq",  "%f",  " %10.1f",      "0",    "0",   4,  0, 0x003, 0, "Horizontale Ausdehnung der Quelle (m)" },
 { "cq",  "%f",  " %10.1f",      "0",    "0",   4,  0, 0x003, 0, "Vertikale Ausdehnung der Quelle (m)" },
 { "eq",  "%f",  " %10.1f",      "1",    "0",   4,  0, 0x003, 0, "Quellstrke (1/s)" },
 { "hq",  "%f",  " %10.1f",    "100",    "0",   4,  0, 0x003, 0, "Hhe der Quelle/Unterkante (m)" },
 { "pq",  "%f",  " %10.1f",      "0",    "0",   4,  0, 0x003, 0, "Winkel von aq gegen Ost gegen den Uhrzeiger (Grad)" },
 { "sq",  "%f",  " %10.2f",      "0",    "0",   4,  0, 0x003, 0, "Anstieg von aq (1)" },
 { "uu",  "%f",  " %10.1f",      "0",    "3",   4,  3, 0x103, 0, "berhhung: Anfangsgeschwindigkeit (m/s)" },
 { "ud",  "%f",  " %10.1f",      "0",    "0",   4,  0, 0x003, 0, "berhhung: Streuintensitt (1)" },
 { "ut",  "%f",  " %10.1f",      "0",    "0",   4,  0, 0x003, 0, "berhhung: Zeitkonstante (s)" },
 { "xq",  "%f",  " %10.0f",      "0",    "0",   4,  0, 0x003, 0, "X-Koordinate der Quelle/Eckpunkt (m)" },
 { "yq",  "%f",  " %10.0f",      "0",    "0",   4,  0, 0x003, 0, "Y-Koordinate der Quelle/Eckpunkt (m)" },
 { "*r",  "%s",  " \"%s\"",      "0",    "0", 255,  0, 0x005, 0, "Read bp|bm|vx|vy|vs|vv|su|sv|sw|sc|ku|kv|kw|vd|pd|li" },
 { "fn",  "%s",  " \"%s\"",       "",    "0", 255,  0, 0x005, 0, "Dateiname" },
 { "rc",  "%d",    " %10d",     "-1",    "0",   4,  0, 0x002, 0, "Lese-Zhler" },
 { "*w",  "%s",  " \"%s\"",       "",    "0", 255,  0, 0x005, 0, "Write cnc|icp|dry|wet|wnd|drf|sgm|lmb|psi|eps|pdp|tau" },
 { "aw",  "%d",    " %10d",      "0",    "0",   4,  0, 0x002, 0, "Schreib-Zhler Offset" },
 { "fa",  "%f",  " %10.1e",    "1.0",    "0",   4,  0, 0x003, 0, "Skalierungsfaktor" },
 { "fi",  "%s",  " \"%s\"",       "",    "0", 255,  0, 0x005, 0, "Dateiname" },
 { "fo",  "%s",  " \"%s\"", "%12.4e",    "0", 255,  0, 0x005, 0, "Format fr die Datenelemente" },
 { "id",  "%s",  " \"%s\"",   "TEST",    "0", 255,  0, 0x005, 0, "Beschreibung" },
 { "mo",  "%s",  " \"%s\"",   "text",    "0", 255,  0, 0x005, 0, "Modus fr die Daten-Reprsentation" },
 { "se",  "%s",  " \"%s\"",       "",    "0", 255,  0, 0x005, 0, "Index-Sequenz" },
 { "wc",  "%d",    " %10d",     "-1",    "0",   4,  0, 0x002, 0, "Schreib-Zhler" },
 { "*z",  "%s",  " \"%s\"",       "",    "0", 255,  0, 0x005, 0, "Parameter fr Simulationslauf" },
 { "dt",  "%f",  " %10.0f",   "1000",    "0",   4,  0, 0x003, 0, "Lnge eines Simulationsintervalls (s)" },
 { "*t",  "%s",  " \"%s\"",       "",    "0", 255,  0, 0x005, 0, "Tabellenausgabe in die Log-Datei" },
 { "*l",  "%s",  " \"%s\"",       "",    "0", 255,  0, 0x005, 0, "Beginn der Wiederholungs-Schleife" },
 { "*n",  "%s",  " \"%s\"",       "",    "0", 255,  0, 0x005, 0, "Ende der Wiederholungs-Schleife" },
 { "*e",  "%s",  " \"%s\"",       "",    "0", 255,  0, 0x005, 0, "Ende der Rechnung" },
 { "**", NULL, NULL, NULL, NULL, 0, 0, 0, NULL } }
#endif
;
//  Macros for easier reference:

#define  MC  PrmTab.mc.u.i
#define  MX  PrmTab.mx.u.i
#define  MY  PrmTab.my.u.i
#define  MZ  PrmTab.mz.u.i
#define  NX  PrmTab.nx.u.i
#define  NY  PrmTab.ny.u.i
#define  NZ  PrmTab.nz.u.i
#define  OP  (char*)PrmTab.op.u.p
#define  SD  PrmTab.sd.u.i
#define  TI  (char*)PrmTab.ti.u.p
#define  A0  PrmTab.a0.u.f
#define  B0  PrmTab.b0.u.f
#define CC(i) ((float*)PrmTab.cc.u.p)[i]
#define  DA  PrmTab.da.u.f
#define  DC  PrmTab.dc.u.f
#define  DD  PrmTab.dd.u.f
#define  SH  PrmTab.sh.u.f
#define  X0  PrmTab.x0.u.f
#define  X1  PrmTab.x1.u.f
#define  Y0  PrmTab.y0.u.f
#define  Y1  PrmTab.y1.u.f
#define  ZH  PrmTab.zh.u.f
#define ZZ(i) ((float*)PrmTab.zz.u.p)[i]
#define  AA  PrmTab.aa.u.f
#define KU(i) ((float*)PrmTab.ku.u.p)[i]
#define KV(i) ((float*)PrmTab.kv.u.p)[i]
#define KW(i) ((float*)PrmTab.kw.u.p)[i]
#define  PD  PrmTab.pd.u.f
#define  QP  PrmTab.qp.u.f
#define  RP  PrmTab.rp.u.f
#define  RW  PrmTab.rw.u.f
#define SC(i) ((float*)PrmTab.sc.u.p)[i]
#define SU(i) ((float*)PrmTab.su.u.p)[i]
#define SV(i) ((float*)PrmTab.sv.u.p)[i]
#define SW(i) ((float*)PrmTab.sw.u.p)[i]
#define  TA  PrmTab.ta.u.f
#define TT(i) ((float*)PrmTab.tt.u.p)[i]
#define TU(i) ((float*)PrmTab.tu.u.p)[i]
#define TV(i) ((float*)PrmTab.tv.u.p)[i]
#define TW(i) ((float*)PrmTab.tw.u.p)[i]
#define  VD  PrmTab.vd.u.f
#define  VS  PrmTab.vs.u.f
#define VX(i) ((float*)PrmTab.vx.u.p)[i]
#define VY(i) ((float*)PrmTab.vy.u.p)[i]
#define  AQ  PrmTab.aq.u.f
#define  BQ  PrmTab.bq.u.f
#define  CQ  PrmTab.cq.u.f
#define  EQ  PrmTab.eq.u.f
#define  HQ  PrmTab.hq.u.f
#define  PQ  PrmTab.pq.u.f
#define  SQ  PrmTab.sq.u.f
#define UU(i) ((float*)PrmTab.uu.u.p)[i]
#define  UD  PrmTab.ud.u.f
#define  UT  PrmTab.ut.u.f
#define  XQ  PrmTab.xq.u.f
#define  YQ  PrmTab.yq.u.f
#define  FN  (char*)PrmTab.fn.u.p
#define  RC  PrmTab.rc.u.i
#define  AW  PrmTab.aw.u.i
#define  FA  PrmTab.fa.u.f
#define  FI  (char*)PrmTab.fi.u.p
#define  FO  (char*)PrmTab.fo.u.p
#define  ID  (char*)PrmTab.id.u.p
#define  MO  (char*)PrmTab.mo.u.p
#define  SE  (char*)PrmTab.se.u.p
#define  WC  PrmTab.wc.u.i
#define  DT  PrmTab.dt.u.f
